# Copyright (c) 2015, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

# stdlib
import collections
import json
import warnings

from . import namespaces
from . import idgen, signals
from .binding_utils import save_encoding
from .datautils import is_sequence
from .vendor import six

# Note: Some imports moved to the bottom of this module to resolve circular
# import issues.


def _objectify(field, value, ns_info):
    """Make `value` suitable for a binding object.

    If `value` is an Entity, call to_obj() on it. Otherwise, pass it
    off to the TypedField for an appropriate value.
    """
    if (getattr(field.type_, "_treat_none_as_empty_list", False) and
            value is None):
        return []

    if value is None:
        return None
    elif field.type_:
        return value.to_obj(ns_info=ns_info)
    return field.binding_value(value)


def _dictify(field, value):
    """Make `value` suitable for a dictionary.

    * If `value` is an Entity, call to_dict() on it.
    * If value is a timestamp, turn it into a string value.
    * If none of the above are satisfied, return the input value
    """
    if value is None:
        return None
    elif field.type_:
        return value.to_dict()
    return field.dict_value(value)


class EntityFactory(object):
    """The EntityFactory class handles type key => class resolution during
    parse time.

    Generally, this class will use xsi:type as a class lookup key, but there
    may be instances where xsi:type is not be the appropriate key. As such
    this class was designed to be flexible enough to handle multiple class
    resolution scenarios.

    Any of the methods defined below can be overridden for non-standard
    keys or lookup scenarios.

    Attributes:
        _dictkey: The key associated with type information in an Entity
            dictionary. Default is "xsi:type".
        _objkey: The attribute associated with type information in a
            generateDS binding object. Default is "xsi_type"
    """

    _dictkey = "xsi:type"  # dictionary key holding type information
    _objkey  = "xsi_type"  # generateDS object attr holding type information.


    @classmethod
    def entity_class(cls, key):
        """**Abstract** This method take the `key` (e.g., xsi:type value) and
        returns the Entity class associated with it.

        Args:
            key: The class lookup key. Generally this is an xsi:type attribute
                value.

        Returns:
            The class associated with the input `key`.

        Raises:
            This should raise ValueError if no class is associated with the
            input key.
        """
        raise NotImplementedError()

    @classmethod
    def instance(cls, key, *args, **kwargs):
        """Create an instance of the class associated with the `key` (xsi:type)
        and initialize it with the *args and **kwargs.

        Args:
            key: A class lookup key (see entity_class()).

        Returns:
            An instance of the class associated with the `key`.
        """
        klass = cls.entity_class(key)
        return klass(*args, **kwargs)

    @classmethod
    def objkey(cls, obj):
        """Return the class lookup key from the input generateDS object.
        If no type information or lookup keys are found on the binding object
        return None.

        Args:
            obj: A generateDS binding object.

        Returns:
            A class lookup key (string) or None.
        """
        return getattr(obj, cls._objkey, None)

    @classmethod
    def dictkey(cls, mapping):
        """Return the class lookup key from the input dictionary.
        If no type information or lookup keys are found on the binding object
        return None.

        Args:
            mapping: A dictionary representation of an Entity instance.

        Returns:
            A class lookup key (string) or None.
        """
        try:
            return mapping[cls._dictkey]
        except (KeyError, AttributeError):
            return None

    @classmethod
    def from_dict(cls, cls_dict, fallback_xsi_type=None):
        """Parse the dictionary and return an Entity instance.

        This will attempt to extract type information from the input
        dictionary and pass it to entity_class to resolve the correct class
        for the type.

        Args:
            cls_dict: A dictionary representation of an Entity object.
            fallback_xsi_type: An xsi_type to use for string input, which doesn't have properties

        Returns:
            An Entity instance.
        """
        if not cls_dict:
            return None
        
        if isinstance(cls_dict, six.string_types):
            if not getattr(cls, "_convert_strings", False):
                return cls_dict

        try:
            typekey = cls.dictkey(cls_dict)
        except TypeError:
            typekey = fallback_xsi_type
        klass   = cls.entity_class(typekey)
        return klass.from_dict(cls_dict)

    @classmethod
    def from_obj(cls, cls_obj):
        """Parse the generateDS object and return an Entity instance.

        This will attempt to extract type information from the input
        object and pass it to entity_class to resolve the correct class
        for the type.

        Args:
            cls_obj: A generateDS object.

        Returns:
            An Entity instance.
        """
        if not cls_obj:
            return None

        typekey = cls.objkey(cls_obj)
        klass   = cls.entity_class(typekey)
        return klass.from_obj(cls_obj)


class Entity(object):
    """Base class for all classes in the Cybox SimpleAPI."""

    # By default (unless a particular subclass states otherwise), try to "cast"
    # invalid objects to the correct class using the constructor. Entity
    # subclasses should either provide a "sane" constructor or set this to
    # False.
    _try_cast = True

    def __init__(self):
        self._fields = {}

    @classmethod
    def typed_fields(cls):
        """Return a tuple of this entity's TypedFields."""

        # Checking cls._typed_fields could return a superclass _typed_fields
        # value. So we check our class __dict__ which does not include
        # inherited attributes.
        klassdict = cls.__dict__

        try:
            return klassdict["_typed_fields"]
        except KeyError:
            fields = cls.typed_fields_with_attrnames()
            cls._typed_fields = tuple(field for _, field in fields)
        return cls._typed_fields

    @classmethod
    def typed_fields_with_attrnames(cls):
        """Return a list of (TypedField attribute name, TypedField object)
        tuples for this Entity.
        """
        # Checking cls._typed_fields could return a superclass _typed_fields
        # value. So we check our class __dict__ which does not include
        # inherited attributes.
        klassdict = cls.__dict__

        try:
            return klassdict["_typed_fields_with_attrnames"]
        except KeyError:
            # No typed_fields set on this Entity yet. Find them and store
            # them in the _typed_fields class attribute.
            from . import fields
            typed_fields = tuple(fields.iterfields(cls))
            cls._typed_fields_with_attrnames = typed_fields
        return typed_fields

    def __eq__(self, other):
        # This fixes some strange behavior where an object isn't equal to
        # itself
        if other is self:
            return True

        # I'm not sure about this, if we want to compare exact classes or if
        # various subclasses will also do (I think not), but for now I'm going
        # to assume they must be equal. - GTB
        if self.__class__ != other.__class__:
            return False

        # Get all comparable TypedFields
        typedfields = [f for f in self.typed_fields() if f.comparable]

        # If No comparable TypedFields are found, return False so we don't
        # inadvertantly say they are equal.
        if not typedfields:
            return False

        return all(f.__get__(self) == f.__get__(other) for f in typedfields)

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_obj(self, ns_info=None):
        """Convert to a GenerateDS binding object.

        Subclasses can override this function.

        Returns:
            An instance of this Entity's ``_binding_class`` with properties
            set from this Entity.
        """
        if ns_info:
            ns_info.collect(self)

        # null behavior for classes that inherit from Entity but do not have _binding_class
        if not hasattr(self, "_binding_class"):
            return None

        entity_obj = self._binding_class()

        for field, val in six.iteritems(self._fields):
            # EntityLists with no list items should be dropped
            if isinstance(val, EntityList) and len(val)==0:
                val = None
            elif field.multiple:
                if val:
                    val = [_objectify(field, x, ns_info) for x in val]
                else:
                    val = []
            else:
                val = _objectify(field, val, ns_info)

            setattr(entity_obj, field.name, val)

        self._finalize_obj(entity_obj)
        return entity_obj

    def _finalize_obj(self, entity_obj):
        """Subclasses can define additional items in the binding object.

        `entity_obj` should be modified in place.
        """
        pass

    def to_dict(self):
        """Convert to a ``dict``

        Subclasses can override this function.

        Returns:
            Python dict with keys set from this Entity.
        """
        entity_dict = {}

        for field, val in six.iteritems(self._fields):
            if field.multiple:
                if val:
                    val = [_dictify(field, x) for x in val]
                else:
                    val = []
            else:
                val = _dictify(field, val)

            # Only add non-None objects or non-empty lists
            if val is not None and val != []:
                entity_dict[field.key_name] = val

        self._finalize_dict(entity_dict)

        return entity_dict

    def _finalize_dict(self, entity_dict):
        """Subclasses can define additional items in the dictionary.

        `entity_dict` should be modified in place.
        """
        pass

    @classmethod
    def from_obj(cls, cls_obj):
        if not cls_obj:
            return None

        entity = cls()

        for field in cls.typed_fields():
            try:
                val = getattr(cls_obj, field.name)
            except:
                # HACK: if the cls_obj is not an instance of cls,
                # only copy fields that exist on cls
                # (this only happens with "Base" binding types that map onto
                #  non-base API types, like BaseIndicatorType => Indicator)
                continue

            # Get the class that will perform the from_obj() call and
            # transform the generateDS binding object into an Entity.
            transformer = field.transformer

            if transformer:
                if field.multiple and val is not None:
                    val = [transformer.from_obj(x) for x in val]
                else:
                    val = transformer.from_obj(val)

            field.__set__(entity, val)

        signals.emit("Entity.created.from_obj", entity, cls_obj)
        return entity

    @classmethod
    def from_dict(cls, cls_dict):
        if cls_dict is None:
            return None

        entity = cls()

        # Shortcut if an actual dict is not provided:
        if not isinstance(cls_dict, dict):
            value = cls_dict

            try:
                return cls(value)   # Call the class's constructor
            except TypeError as ex:
                fmt  = "Could not instantiate a %s from a %s: %s"
                args = (cls, type(value), value)
                ex.message = fmt % args
                raise

        for field in cls.typed_fields():
            val = cls_dict.get(field.key_name)

            # Get the class that will perform the from_obj() call and
            # transform the generateDS binding object into an Entity.
            transformer = field.transformer
            if transformer:
                if field.multiple:
                    if val is not None:
                        val = [transformer.from_dict(x) for x in val]
                    else:
                        val = []
                else:
                    val = transformer.from_dict(val)
            elif field.multiple and not val:
                val = []

            # Set the value
            field.__set__(entity, val)

        return entity

    def to_xml(self, include_namespaces=True, namespace_dict=None,
               pretty=True, encoding='utf-8'):
        """Serializes a :class:`Entity` instance to an XML string.

        The default character encoding is ``utf-8`` and can be set via the
        `encoding` parameter. If `encoding` is ``None``, a unicode string
        is returned.

        Args:
            include_namespaces (bool): whether to include xmlns and
                xsi:schemaLocation attributes on the root element. Set to true by
                default.
            namespace_dict (dict): mapping of additional XML namespaces to
                prefixes
            pretty (bool): whether to produce readable (``True``) or compact
                (``False``) output. Defaults to ``True``.
            encoding: The output character encoding. Default is ``utf-8``. If
                `encoding` is set to ``None``, a unicode string is returned.

        Returns:
            An XML string for this
            :class:`Entity` instance. Default character encoding is ``utf-8``.

        """
        namespace_def = ""

        ns_collector = NamespaceCollector()
        gds_obj = self.to_obj(ns_info=ns_collector if include_namespaces else None)

        if include_namespaces:
            ns_collector.finalize(namespace_dict)
            delim = "\n\t" if pretty else " "
            namespace_def = ns_collector.get_xmlns_string(delim) + delim + \
                ns_collector.get_schema_location_string(delim)

        with save_encoding(encoding):
            sio = six.StringIO()
            gds_obj.export(
                sio.write,
                0,
                namespacedef_=namespace_def,
                pretty_print=pretty
            )

        s = six.text_type(sio.getvalue()).strip()

        if encoding:
            return s.encode(encoding)

        return s

    def to_json(self):
        """Export an object as a JSON String."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_doc):
        """Parse a JSON string and build an entity."""
        try:
            d = json.load(json_doc)
        except AttributeError: # catch the read() error
            d = json.loads(json_doc)

        return cls.from_dict(d)

    @classmethod
    def istypeof(cls, obj):
        """Check if `cls` is the type of `obj`

        In the normal case, as implemented here, a simple isinstance check is
        used. However, there are more complex checks possible. For instance,
        EmailAddress.istypeof(obj) checks if obj is an Address object with
        a category of Address.CAT_EMAIL
        """
        return isinstance(obj, cls)

    @classmethod
    def object_from_dict(cls, entity_dict):
        """Convert from dict representation to object representation."""
        return cls.from_dict(entity_dict).to_obj()

    @classmethod
    def dict_from_object(cls, entity_obj):
        """Convert from object representation to dict representation."""
        return cls.from_obj(entity_obj).to_dict()


class EntityList(collections.MutableSequence, Entity):
    """An EntityList is an Entity that behaves like a mutable sequence.

    EntityList implementations must define one multiple TypedField which
    has an Entity subclass type. EntityLists can define other TypedFields
    that are not multiple.

    The MutableSequence methods are used to interact with the multiple
    TypedField.
    """

    # Don't try to cast list types (yet)
    _try_cast = False


    def __init__(self, *args):
        super(EntityList, self).__init__()
        assert self._multiple_field()

        if not any(args):
            return

        for arg in args:
            if is_sequence(arg):
                self.extend(arg)
            else:
                self.append(arg)

    def _any_typedfields(self):
        return any(x for x in six.itervalues(self._fields))

    def __nonzero__(self):
        return bool(self._inner or self._any_typedfields())

    __bool__ = __nonzero__

    def __getitem__(self, key):
        return self._inner.__getitem__(key)

    def __setitem__(self, key, value):
        self._inner.__setitem__(key, value)

    def __delitem__(self, key):
        self._inner.__delitem__(key)

    def __len__(self):
        return self._inner.__len__()

    def insert(self, idx, value):
        if not value:
            return
        self._inner.insert(idx, value)

    @classmethod
    def _dict_as_list(cls):
        """Returns True if the to_dict() and from_dict() methods should
        expect to export/parse lists rather than dicts.

        If there is only one TypedField (a multiple field) defined on the
        EntityList, this will return True. If there is more than one
        TypedField, the to_dict() method will export a dict and not a list.

        Note:
            This can be overridden by subclasses to force this behavior.

        TODO (bworrell):
            Remove this functionality. The to_dict() method can export a
            dict, a list, a string/int/etc., or None. This makes parsing
            complicated. The to_dict() method should probably return a dict
            and maybe None at most.
        """
        return len(cls.typed_fields()) == 1

    @classmethod
    def _multiple_field(cls):
        """Return the "multiple" TypedField associated with this EntityList.

        This also lazily sets the ``_entitylist_multiplefield`` value if it
        hasn't been set yet. This is set to a tuple containing one item because
        if we set the class attribute to the TypedField, we would effectively
        add a TypedField descriptor to the class, which we don't want.

        Raises:
            AssertionError: If there is more than one multiple TypedField
                or the the TypedField type_ is not a subclass of Entity.
        """
        klassdict = cls.__dict__

        try:
            # Checking for cls.entitylist_multifield would return any inherited
            # values, so we check the class __dict__ explicitly.
            return klassdict["_entitylist_multifield"][0]
        except (KeyError, IndexError, TypeError):
            from . import fields
            multifield_tuple = tuple(fields.find(cls, multiple=True))
            assert len(multifield_tuple) == 1

            # Make sure that the multiple field actually has an Entity type.
            multifield  = multifield_tuple[0]
            assert issubclass(multifield.type_, Entity)

            # Store aside the multiple field. We wrap it in a tuple because
            # just doing ``cls._entitylist_multifield = multifield`` would
            # assign another TypedField descriptor to this class. We don't
            # want that.
            cls._entitylist_multifield =  multifield_tuple

            # Return the multiple TypedField
            return multifield_tuple[0]

    @property
    def _inner(self):
        """Return the EntityList mutliple TypedField inner collection."""
        return self._multiple_field().__get__(self)

    def to_list(self):
        return [h.to_dict() for h in self]

    def to_dict(self):
        if self._dict_as_list():
            return self.to_list()
        return super(EntityList, self).to_dict()

    @classmethod
    def from_dict(cls, cls_dict):
        if not cls_dict:
            return None

        if cls._dict_as_list():
            return cls.from_list(cls_dict)

        return super(EntityList, cls).from_dict(cls_dict)

    @classmethod
    def from_list(cls, seq):
        if not seq:
            return None

        entitylist  = cls()
        transformer = cls._multiple_field().transformer
        fallback_xsi_type = getattr(cls._multiple_field().type_,"_XSI_TYPE", None)
        try:
            transformed_list = [transformer.from_dict(x, fallback_xsi_type)
                                for x in seq]
        except TypeError:
            transformed_list = [transformer.from_dict(x)
                                for x in seq]
        entitylist.extend(transformed_list)

        return entitylist

    @classmethod
    def object_from_list(cls, entitylist_list):
        """Convert from list representation to object representation."""
        return cls.from_list(entitylist_list).to_obj()

    @classmethod
    def list_from_object(cls, entitylist_obj):
        """Convert from object representation to list representation."""
        return cls.from_obj(entitylist_obj).to_list()


class NamespaceCollector(object):

    def __init__(self):
        # Namespaces that are "collected" from the Python objects during
        # serialization.  This will be a NamespaceSet.
        self._collected_namespaces = None

        # Namespaces and schemalocations that are attached to STIX/CybOX
        # entities when parsed from an external source.
        self._input_namespaces = {}
        self._input_schemalocs = {}

        # A list of classes that have been visited/seen during the namespace
        # collection process. This speeds up the collect() method.
        self._collected_classes = set()

        # Namespaces and schemalocations that will appear in the output
        # XML document.
        self.finalized_schemalocs = None

        # Namespace dictionary that gets passed to the bindings.
        self.binding_namespaces = None

    def update(self, ns_info):
        self._collected_namespaces.update(ns_info._collected_namespaces)  # noqa
        self._input_namespaces.update(ns_info._input_namespaces)  # noqa
        self._input_schemalocs.update(ns_info._input_schemalocs)  # noqa

    def _parse_collected_classes(self):
        alias2uri = {}     # namespace alias to namespace uri mapping
        noalias   = set()  # namespaces that have no alias defined by a class.

        for klass in self._collected_classes:
            # Prevents exception being raised if/when
            # collections.MutableSequence or another base class appears in the
            # MRO.
            ns = getattr(klass, "_namespace", None)
            if not ns:
                continue

            # cybox.objects.* ObjectProperties derivations have an _XSI_NS
            # class-level attribute which holds the namespace alias to be
            # used for its namespace.
            alias = getattr(klass, "_XSI_NS", None)
            if alias:
                alias2uri[alias] = ns
                continue

            # Many stix/cybox entity classes have an _XSI_TYPE attribute that
            # contains a `prefix:namespace` formatted QNAME for the
            # associated xsi:type.
            xsi_type = getattr(klass, "_XSI_TYPE", None)
            if not xsi_type:
                noalias.add(ns)
                continue

            # Attempt to split the xsi:type attribute value into the ns alias
            # and the typename.
            typeinfo = xsi_type.split(":")
            if len(typeinfo) == 2:
                alias2uri[typeinfo[0]] = ns
            else:
                noalias.add(ns)

        # Compile a list of all seen Namespace URIs
        uris = list(six.itervalues(alias2uri)) + list(noalias)

        # A mixbox NamespaceSet object from our collected URIs.
        # Unrecognized namespace URIs will cause an error at this stage.
        nsset = namespaces.make_namespace_subset_from_uris(uris)

        # For some reason, prefixes are specified in API class vars and also in
        # our big namespace tables.  From python-cybox issue #274 [1], I
        # conclude that the tables may take priority here.  And those are
        # already going to be preferred at this point.  So the only thing I can
        # think to do with class var values is fill in any missing prefixes
        # we may have (but I doubt there will be any).
        #
        # 1. https://github.com/CybOXProject/python-cybox/issues/274
        for prefix, ns_uri in six.iteritems(alias2uri):
            preferred_prefix = nsset.preferred_prefix_for_namespace(ns_uri)

            if preferred_prefix:
                continue

            nsset.set_preferred_prefix_for_namespace(
                ns_uri=ns_uri,
                prefix=prefix,
                add_if_not_exist=True
            )

        # Set our internal _collected_namespaces to the populated NamespaceSet
        self._collected_namespaces = nsset

    def _fix_example_namespace(self):
        """Attempts to resolve issues where our samples use
        'http://example.com/' for our example namespace but python-stix uses
        'http://example.com' by removing the former.
        """
        example_prefix = 'example'  # Example ns prefix
        idgen_prefix = idgen.get_id_namespace_prefix()

        # If the ID namespace alias doesn't match the example alias, return.
        if idgen_prefix != example_prefix:
            return

        # If the example namespace prefix isn't in the parsed namespace
        # prefixes, return.
        if example_prefix not in self._input_namespaces:
            return

        self._input_namespaces[example_prefix] = idgen.EXAMPLE_NAMESPACE.name

    def _finalize_namespaces(self, ns_dict=None):
        """Returns a dictionary of namespaces to be exported with an XML
        document.

        This loops over all the namespaces that were discovered and built
        during the execution of ``collect()`` and
        ``_parse_collected_classes()`` and attempts to merge them all.

        Raises:
            .namespaces.DuplicatePrefixError: If namespace prefix was
                mapped to more than one namespace.
            .namespaces.NoPrefixError: If a namespace was collected that is
                not mapped to a prefix.
        """
        if ns_dict:
            # Add the user's entries to our set
            for ns, alias in six.iteritems(ns_dict):
                self._collected_namespaces.add_namespace_uri(ns, alias)

        # Add the ID namespaces
        self._collected_namespaces.add_namespace_uri(
            ns_uri=idgen.get_id_namespace(),
            prefix=idgen.get_id_namespace_alias()
        )

        # Remap the example namespace to the one expected by the APIs if the
        # sample example namespace is found.
        self._fix_example_namespace()

        # Add _input_namespaces
        for prefix, uri in six.iteritems(self._input_namespaces):
            self._collected_namespaces.add_namespace_uri(uri, prefix)

        # Add some default XML namespaces to make sure they're there.
        self._collected_namespaces.import_from(namespaces.XML_NAMESPACES)

        # python-stix's generateDS-generated binding classes can't handle
        # default namespaces.  So make sure there are no preferred defaults in
        # the set.  Get prefixes from the global namespace set if we have to.
        for ns_uri in self._collected_namespaces.namespace_uris:
            preferred_prefix = self._collected_namespaces.preferred_prefix_for_namespace(ns_uri)

            if preferred_prefix:
                continue

            # No preferred prefix set for namespace. Try to assign one.
            prefixes = self._collected_namespaces.get_prefixes(ns_uri)

            if prefixes:
                prefix = next(iter(prefixes))
            else:
                prefix = namespaces.lookup_name(ns_uri)

            if prefix is None:
                raise namespaces.NoPrefixesError(ns_uri)

            self._collected_namespaces.set_preferred_prefix_for_namespace(
                ns_uri=ns_uri,
                prefix=prefix,
                add_if_not_exist=True
            )

    def _finalize_schemalocs(self, schemaloc_dict=None):

        # If schemaloc_dict was passed in, make a copy so we don't mistakenly
        # modify the original.
        if schemaloc_dict:
            schemaloc_dict = schemaloc_dict.copy()
        else:
            schemaloc_dict = {}

        # Build our schemalocation dictionary!
        #
        # Initialize it from values found in the parsed, input schemalocations
        # (if there are any) and the schemaloc_dict parameter values (if there
        # are any).
        #
        # If there is a schemalocation found in both the parsed schemalocs and
        # the schema_loc dict, use the schemaloc_dict value.
        for ns, loc in six.iteritems(self._input_schemalocs):
            if ns not in schemaloc_dict:
                schemaloc_dict[ns] = loc

        # Now use the merged dict to update any schema locations we don't
        # already have.
        for ns, loc in six.iteritems(schemaloc_dict):
            if (ns in self._collected_namespaces and
                self._collected_namespaces.get_schema_location(ns) is None):
                self._collected_namespaces.set_schema_location(ns, loc)

        # Warn if we are missing any schemalocations
        id_ns = idgen.get_id_namespace()

        for ns in self._collected_namespaces.namespace_uris:
            if self._collected_namespaces.get_schema_location(ns):
                continue

            if (ns == id_ns or
                ns in namespaces.XML_NAMESPACES or
                ns in schemaloc_dict):
                continue

            error = "Unable to map namespace '{0}' to schemaLocation"
            warnings.warn(error.format(ns))

    def finalize(self, ns_dict=None, schemaloc_dict=None):
        self._parse_collected_classes()
        self._finalize_namespaces(ns_dict)
        self._finalize_schemalocs(schemaloc_dict)

        self.finalized_schemalocs = self._collected_namespaces.get_uri_schemaloc_map()
        self.binding_namespaces = self._collected_namespaces.get_uri_prefix_map()

    def get_xmlns_string(self, delim):
        if self._collected_namespaces is None:
            return ""

        return self._collected_namespaces.get_xmlns_string(
            preferred_prefixes_only=False,
            delim=delim
        )

    def get_schema_location_string(self, delim):
        if self._collected_namespaces is None:
            return ""
        return self._collected_namespaces.get_schemaloc_string(delim=delim)

    def collect(self, entity):
        # Collect all the classes we need to inspect for namespace information
        self._collected_classes.update(entity.__class__.__mro__)

        # Collect the input namespaces if this entity came from some external
        # source.
        if hasattr(entity, "__input_namespaces__"):
            self._input_namespaces.update(entity.__input_namespaces__)

        # Collect the input schemalocation information if this entity came
        # from some external source.
        if hasattr(entity, "__input_schemalocations__"):
            self._input_schemalocs.update(entity.__input_schemalocations__)



# Copyright (c) 2015, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.
import collections
import sys

from .datautils import is_sequence, resolve_class, needkwargs
from .entities import EntityList
from .vendor import six


class TypedList(collections.MutableSequence):
    """A type-aware mutable sequence that performs input validation when
    inserting new items.

    Args:
        type: The type of the items contained in this collection.
        ignore_none: If True, do not insert None values.
        castfunc: A callable that will convert non-valid items into
            valid items.
        *args: A variable-length list of items to add to the collection.
            If an arg is a non-string, non-EntityList iterable type, each of
            its contained items will be added.
    """

    @needkwargs("type")
    def __init__(self, *args, **kwargs):
        self._inner       = []
        self._type        = resolve_class(kwargs["type"])
        self._castfunc    = kwargs.get("castfunc", self._type)
        self._ignore_none = kwargs.get("ignore_none", True)

        for item in args:
            if isinstance(item, EntityList):
                self.append(item)
            elif is_sequence(item):
                self.extend(item)
            else:
                self.append(item)

    def _is_valid(self, value):
        """Return True if the input value is valid for insertion into the
        inner list.

        Args:
            value: An object about to be inserted.
        """

        # Entities have an istypeof method that can perform more sophisticated
        # type checking.
        if hasattr(self._type, "istypeof"):
            return self._type.istypeof(value)
        else:
            return isinstance(value, self._type)

    def _fix_value(self, value):
        """Attempt to coerce value into the correct type.

        Subclasses can override this function.
        """
        try:
            return self._castfunc(value)
        except:
            error = "Can't put '{0}' ({1}) into a {2}. Expected a {3} object."
            error = error.format(
                value,                  # Input value
                type(value),            # Type of input value
                type(self),             # Type of collection
                self._type              # Expected type of input value
            )
            six.reraise(TypeError, TypeError(error), sys.exc_info()[-1])

    def _is_type_castable(self):
        return getattr(self._type, "_try_cast", False)

    def __nonzero__(self):
        return bool(self._inner)

    def __getitem__(self, key):
        return self._inner.__getitem__(key)

    def __setitem__(self, key, value):
        """Attempt to set the value at position `key` to the `value`.

        If a value is not the correct type, an attempt will be made to
        convert it to the correct type.

        Args:
            key: An index.
            value: A value to set.
        """
        if not self._is_valid(value):
            value = self._fix_value(value)
        self._inner.__setitem__(key, value)

    def __delitem__(self, key):
        self._inner.__delitem__(key)

    def __len__(self):
        return self._inner.__len__()

    def insert(self, idx, value):
        if value is None and self._ignore_none:
            return
        elif self._is_valid(value):
            self._inner.insert(idx, value)
        elif self._is_type_castable():
            value = self._fix_value(value)
            self._inner.insert(idx, value)
        else:
            err = "Cannot insert type (%s) into %s" % (type(value), type(self))
            raise TypeError(err)

    def __repr__(self):
        return self._inner.__repr__()

    def __str__(self):
        return self._inner.__str__()



# Copyright (c) 2015, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

from abc import ABCMeta, abstractmethod
from distutils.version import StrictVersion

from .exceptions import ignored
from .xml import get_etree_root, get_etree, get_schemaloc_pairs
from .vendor.six import iteritems


class UnknownVersionError(Exception):
    """A parsed document contains no version information."""
    pass


class UnsupportedVersionError(Exception):
    """A parsed document is a version unsupported by the parser."""

    def __init__(self, message, expected=None, found=None):
        super(UnsupportedVersionError, self).__init__(message)
        self.expected = expected
        self.found = found


class UnsupportedRootElementError(Exception):
    """A parsed document contains an unsupported root element."""

    def __init__(self, message, expected=None, found=None):
        super(UnsupportedRootElementError, self).__init__(message)
        self.expected = expected
        self.found = found


class EntityParser(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def supported_tags(self):
        """Return an iterable of supported document root tags (strings)."""

    @abstractmethod
    def get_version(self, root):
        """Return as a string the schema version used by the document root."""

    @abstractmethod
    def supported_versions(self, tag):
        """Return all the supported versions for a given tag."""

    @abstractmethod
    def get_entity_class(self, tag):
        """Return the class to be returned as the result of parsing."""

    def _get_version(self, root):
        """Return the version of the root element passed in.

        Args:
            root (etree.Element)

        Returns:
            distutils.StrictVersion

        Raises:
            UnknownVersionError
        """
        # Note: STIX and MAEC use a "version" attribute. To support CybOX, a
        # subclass will need to combine "cybox_major_version",
        # "cybox_minor_version", and "cybox_update_version".
        version = self.get_version(root)
        if version:
            return StrictVersion(version)

        raise UnknownVersionError(
            "Unable to determine the version of the input document. No "
            "version information found on the root element."
        )

    def _check_version(self, root):
        """Ensure the root element is a supported version.

        Args:
            root (etree.Element)

        Raises:
            UnsupportedVersionError
        """
        version = self._get_version(root)
        supported = [StrictVersion(x) for x in
                     self.supported_versions(root.tag)]

        if version in supported:
            return

        error = "Document version ({0}) not in supported versions ({1})"
        raise UnsupportedVersionError(
            message=error.format(version, supported),
            expected=supported,
            found=version
        )

    def _check_root_tag(self, root):
        """Check that the XML element tree has a supported root element.

        Args:
            root (etree.Element)

        Raises:
            UnsupportedRootElementError
        """
        supported = self.supported_tags()
        if root.tag in supported:
            return

        error = "Document root element ({0}) not one of ({1})"
        raise UnsupportedRootElementError(
            message=error.format(root.tag, supported),
            expected=supported,
            found=root.tag,
        )

    def parse_xml_to_obj(self, xml_file, check_version=True, check_root=True,
                         encoding=None):
        """Creates a STIX binding object from the supplied xml file.

        Args:
            xml_file: A filename/path or a file-like object representing a STIX
                instance document
            check_version: Inspect the version before parsing.
            check_root: Inspect the root element before parsing.
            encoding: The character encoding of the input `xml_file`.

        Raises:
            .UnknownVersionError: If `check_version` is ``True`` and `xml_file`
                does not contain STIX version information.
            .UnsupportedVersionError: If `check_version` is ``False`` and
                `xml_file` contains an unsupported STIX version.
            .UnsupportedRootElement: If `check_root` is ``True`` and `xml_file`
                contains an invalid root element.

        """
        root = get_etree_root(xml_file, encoding=encoding)

        if check_root:
            self._check_root_tag(root)

        if check_version:
            self._check_version(root)

        entity_class = self.get_entity_class(root.tag)
        entity_obj = entity_class._binding_class.factory()
        entity_obj.build(root)

        return entity_obj

    def parse_xml(self, xml_file, check_version=True, check_root=True,
                  encoding=None):
        """Creates a python-stix STIXPackage object from the supplied xml_file.

        Args:
            xml_file: A filename/path or a file-like object representing a STIX
                instance document
            check_version: Inspect the version before parsing.
            check_root: Inspect the root element before parsing.
            encoding: The character encoding of the input `xml_file`. If
                ``None``, an attempt will be made to determine the input
                character encoding.

        Raises:
            .UnknownVersionError: If `check_version` is ``True`` and `xml_file`
                does not contain STIX version information.
            .UnsupportedVersionError: If `check_version` is ``False`` and
                `xml_file` contains an unsupported STIX version.
            .UnsupportedRootElement: If `check_root` is ``True`` and `xml_file`
                contains an invalid root element.

        """

        xml_etree = get_etree(xml_file, encoding=encoding)
        entity_obj = self.parse_xml_to_obj(
            xml_file=xml_etree,
            check_version=check_version,
            check_root=check_root
        )

        xml_root_node = xml_etree.getroot()
        entity = self.get_entity_class(xml_root_node.tag).from_obj(entity_obj)

        # Save the parsed nsmap and schemalocations onto the parsed Entity
        entity.__input_namespaces__ = dict(iteritems(xml_root_node.nsmap))
        with ignored(KeyError):
            pairs = get_schemaloc_pairs(xml_root_node)
            entity.__input_schemalocations__ = dict(pairs)

        return entity



# Copyright (c) 2015, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.
"""
Utilities for dealing with XML namespaces.
"""

# stdlib
import collections
import copy

# external
from ordered_set import OrderedSet

# internal
from mixbox.vendor import six


_NamespaceTuple = collections.namedtuple("Namespace",
                                         "name prefix schema_location")

class Namespace(_NamespaceTuple):
    """A convenience class which represents simplified XML namespace info,
    consisting of exactly one namespace URI, and an optional prefix and schema
    location URI.  This is handy for building up big tables of namespace data.
    """
    def __new__(cls, name, prefix, schema_location=None):
        """This essentially enables parameter defaulting behavior on a
        namedtuple.  Here, schema location need not be given when creating a
        Namespace, and it will default to None."""
        return super(Namespace, cls).__new__(cls, name, prefix, schema_location)


class DuplicatePrefixError(Exception):
    """Indicates an attempt to map a prefix to two different namespaces."""
    def __init__(self, prefix, *namespaces):
        super(DuplicatePrefixError, self).__init__(
            "Can't map prefix '{0}' to different namespaces: {1}".format(
                prefix, namespaces
            )
        )
        self.prefix = prefix
        self.namespaces = namespaces


class ConflictingSchemaLocationError(Exception):
    """Indicates an attempt to associated an XML namespace URI with two
    different schema location URIs."""
    def __init__(self, ns_uri, *schemalocs):
        super(ConflictingSchemaLocationError, self).__init__(
            "Can't map namespace '{0}' to different schema locations: {1}".format(
                ns_uri, schemalocs
            )
        )
        self.ns_uri = ns_uri
        self.schemalocs = schemalocs


class NamespaceNotFoundError(Exception):
    """Thrown when a namespace is not found.  The URI identifying the namespace
    is available from the "ns_uri" attribute."""
    def __init__(self, ns_uri):
        super(NamespaceNotFoundError, self).__init__(
            "Namespace not found: {0}".format(ns_uri)
        )
        self.ns_uri = ns_uri


class PrefixNotFoundError(Exception):
    """Thrown when a namespace prefix is not found.  The prefix is available
    from the "prefix" attribute."""
    def __init__(self, prefix):
        super(PrefixNotFoundError, self).__init__(
            "Prefix not found: {0}".format(prefix)
        )
        self.prefix = prefix


class TooManyDefaultNamespacesError(Exception):
    """Raised when creating xmlns attributes.  You can't have more than one
    default namespace in an XML document.  If a second preferred default is
    encountered, an attempt is made to choose a prefix from the set of
    registered prefixes.  If none are registered, this exception is raised."""
    def __init__(self, ns_uri):
        super(TooManyDefaultNamespacesError, self).__init__(
            "Too many default namespaces: needed a prefix for namespace '{0}' "
            "but none are defined!".format(ns_uri)
        )


class NoPrefixesError(Exception):
    """Thrown when prefixes are required for a namespace, but none are
    registered."""
    def __init__(self, ns_uri):
        super(NoPrefixesError, self).__init__(
            "Namespace '{0}' has no prefixes!".format(ns_uri)
        )


class InvalidNamespaceSetError(Exception):
    def __init__(self, msg):
        super(InvalidNamespaceSetError, self).__init__(
            "NamespaceSet is invalid: {0}".format(msg)
        )


class _NamespaceInfo(object):
    """**This class is an implementation detail of :class:`NamespaceSet`.
    Others must not use this class.**

    Holds all info about a single XML namespace, including its URI, a
    set of prefixes, and a schema location URI.

    This is an internal class.  Some invariants must be maintained:
    preferred_prefix is a member of prefixes, or is None (meaning the
    preferred thing to do is use it as an XML default namespace).
    There must be no more than one instance per namespace URI.  Other users
    may mess up the invariants, which is why this is "hidden".  The
    NamespaceSet public interface never uses this class.
    """

    def __init__(self, *args):
        """If a Namespace object is passed, this object is constructed from
        it.  If one to three strings are passed, they are treated as
        individual namespace components in the following order: URI, prefix,
        schema location.  Either way, the given prefix will become the
        preferred prefix.
        """
        iterargs = iter(args)
        ns       = next(iterargs, None)
        if ns is None:
            raise ValueError("At least one arg is required")

        if isinstance(ns, Namespace):
            self.__construct_from_namespace(ns)
        else:
            prefix    = next(iterargs, None)
            schemaloc = next(iterargs, None)
            self.__construct_from_components(ns, prefix, schemaloc)

    def __construct_from_namespace(self, ns):
        """Initialize this instance from a given Namespace object."""
        assert isinstance(ns, Namespace)  # TODO (bworrell): Change this to an if not isinstance(...): raise TypeError?
        self.__construct_from_components(*ns)

    def __construct_from_components(self, ns_uri, prefix=None, schema_location=None):
        """Initialize this instance from a namespace URI, and optional
        prefix and schema location URI."""

        assert ns_uri  # other fields are optional

        self.uri = ns_uri
        self.schema_location = schema_location or None
        self.prefixes = OrderedSet()

        if prefix:
            self.prefixes.add(prefix)
        self.preferred_prefix = prefix or None

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return vars(other) == vars(self)

    def __ne__(self, other):
        """Python2 apparently needs this; python3 has a suitable default
        which delegates to __eq__.
        """
        return not self.__eq__(other)

    def __str__(self):
        """for debugging"""
        if self.preferred_prefix:
            preferred_prefix = self.preferred_prefix
        else:
            preferred_prefix = "(default)"

        s = "{0.uri}\n  {0.prefixes}\n  preferred: {1}\n  {0.schema_location}"
        return s.format(self, preferred_prefix)


class NamespaceSet(object):
    """Represents a set of XML namespaces.  For each namespace, a set
    of prefixes and a schema location URI are also maintained.  Prefixes and
    schema location are optional; the namespace URI is always required.

    Each namespace has a preferred prefix.  If None, this indicates a
    preference that it be used as a default namespace.  At present, there is
    nothing preventing multiple namespaces from preferring to be default.  Of
    course, in any given XML document, there can only be one default.  The
    get_xmlns_string() function may throw if there are too many preferred
    default namespaces in this set."""

    def __init__(self):
        # Each mapped-to value in this map must be unique (a _NamespaceInfo).
        self.__ns_uri_map = {}
        # Mapped-to values in this map must refer to a mapped-to value in
        # __ns_uri_map.  More than one key may map to the same value.
        self.__prefix_map = {}

    def __lookup_uri(self, uri):
        """Return a _NamespaceInfo object that is currently associated with the
        input namespace `uri`.
        """
        try:
            return self.__ns_uri_map[uri]
        except KeyError:
            raise NamespaceNotFoundError(uri)

    def __lookup_prefix(self, prefix):
        """Return a _NamespaceInfo object that is currently associated with the
        input `prefix`.
        """
        try:
            return self.__prefix_map[prefix]
        except KeyError:
            raise PrefixNotFoundError(prefix)

    def __add_namespaceinfo(self, ni):
        """Internal method to directly add a _NamespaceInfo object to this
        set.  No sanity checks are done (e.g. checking for prefix conflicts),
        so be sure to do it yourself before calling this."""
        self.__ns_uri_map[ni.uri] = ni
        for prefix in ni.prefixes:
            self.__prefix_map[prefix] = ni

    def __check_prefix_conflict(self, existing_ni_or_ns_uri, incoming_prefix):
        """If existing_ni_or_ns_uri is a _NamespaceInfo object (which must
        be in this set), then caller wants to map incoming_prefix to that
        namespace.  This function verifies that the prefix isn't already mapped
        to a different namespace URI.  If it is, an exception is raised.

        Otherwise, existing_ni_or_ns_uri is treated as a string namespace URI
        which must not already exist in this set.  Caller wants to map
        incoming_prefix to that URI.  If incoming_prefix maps to anything
        already, that represents a prefix conflict and an exception is raised.
        """
        if incoming_prefix not in self.__prefix_map:
            return

        # Prefix found in the prefix map. Check that there are no conflicts
        prefix_check_ni = self.__prefix_map[incoming_prefix]

        if isinstance(existing_ni_or_ns_uri, _NamespaceInfo):
            existing_ni = existing_ni_or_ns_uri  # makes following code clearer?

            if prefix_check_ni is not existing_ni:
                # A different obj implies a different namespace URI is
                # already assigned to the prefix.
                raise DuplicatePrefixError(incoming_prefix, prefix_check_ni.uri, existing_ni.uri)
        else:
            ns_uri = existing_ni_or_ns_uri  # makes following code clearer?
            assert not self.contains_namespace(ns_uri)  # TODO (bworrell): Should this be a raise?
            raise DuplicatePrefixError(incoming_prefix, prefix_check_ni.uri, ns_uri)

    def contains_namespace(self, ns_uri):
        """Determines whether the namespace identified by ns_uri is in this
        set.
        """
        return ns_uri in self.__ns_uri_map

    def __contains__(self, ns_uri):
        """Return True if the `ns_uri` is held by the NamespaceSet.  Just
        invokes contains_namespace().  This method enables the "X in Y" style
        of code to check for containment.

        Args:
            ns_uri: A namespace uri
        """
        return self.contains_namespace(ns_uri)

    def namespace_for_prefix(self, prefix):
        """Get the namespace the given prefix maps to.

        Args:
            prefix (str): The prefix

        Returns:
            str: The namespace, or None if the prefix isn't mapped to
                anything in this set.
        """
        try:
            ni = self.__lookup_prefix(prefix)
        except PrefixNotFoundError:
            return None
        else:
            return ni.uri

    def preferred_prefix_for_namespace(self, ns_uri):
        """Get the "preferred" prefix for the given namespace.  Returns None
        if the preference is to use as the default namespace."""
        ni = self.__lookup_uri(ns_uri)
        return ni.preferred_prefix

    def set_preferred_prefix_for_namespace(self, ns_uri, prefix, add_if_not_exist=False):
        """Sets the preferred prefix for ns_uri.  If add_if_not_exist is True,
        the prefix is added if it's not already registered.  Otherwise,
        setting an unknown prefix as preferred is an error.  The default
        is False.  Setting to None always works, and indicates a preference
        to use the namespace as a default.  The given namespace must already
        be in this set.

        Args:
            ns_uri (str): the namespace URI whose prefix is to be set
            prefix (str): the preferred prefix to set
            add_if_not_exist (bool): Whether to add the prefix if it is not
                already set as a prefix of ``ns_uri``.

        Raises:
            NamespaceNotFoundError: If namespace ``ns_uri`` isn't in this set.
            DuplicatePrefixError: If ``prefix`` already maps to a different
                namespace.
        """
        ni = self.__lookup_uri(ns_uri)

        if not prefix:
            ni.preferred_prefix = None
        elif prefix in ni.prefixes:
            ni.preferred_prefix = prefix
        elif add_if_not_exist:
            self.add_prefix(ns_uri, prefix, set_as_preferred=True)
        else:
            raise PrefixNotFoundError(prefix)

    def __merge_schema_locations(self, ni, incoming_schemaloc):
        """Merge incoming_schemaloc into the given `_NamespaceInfo`, ni.  If we
        don't have one yet and the incoming value is non-None, update ours
        with theirs.  This modifies ni.
        """
        if ni.schema_location == incoming_schemaloc:  # TODO (bworrell): empty strings?
            return
        elif not ni.schema_location:
            ni.schema_location = incoming_schemaloc or None
        elif not incoming_schemaloc:
            return
        else:
            raise ConflictingSchemaLocationError(ni.uri, ni.schema_location, incoming_schemaloc)

    def add_namespace(self, ns):
        """Add a namespace from a :class:`Namespace` object.  This method
        just passes off the tuple fields to :meth:`add_namespace_uri`."""
        assert isinstance(ns, Namespace)
        self.add_namespace_uri(*ns)

    def add_namespace_uri(self, ns_uri, prefix=None, schema_location=None):
        """Adds a new namespace to this set, optionally with a prefix and
        schema location URI.

        If the namespace already exists, the given prefix and schema location
        are merged with the existing entry:
            * If non-None, ``prefix`` is added to the set.  The preferred
                prefix is not modified.
            * If a schema location is not already associated with the
                namespace, it is set to ``schema_location`` (if given).

        If the namespace doesn't already exist in this set (so a new one is
        being created) and a prefix is given, that prefix becomes preferred.
        If not given, a preference as a default namespace is used.

        Args:
            ns_uri (str): The URI of the new namespace
            prefix (str): The desired prefix for the new namespace (optional)
            schema_location (str): The desired schema location for the new
                namespace (optional).

        Raises:
            DuplicatePrefixError: If a prefix is given which already maps to a
                different namespace
            ConflictingSchemaLocationError: If a schema location is given and
                the namespace already exists in this set with a different
                schema location.
        """
        assert ns_uri

        if ns_uri in self.__ns_uri_map:
            # We have a _NamespaceInfo object for this URI already.  So this
            # is a merge operation.
            #
            # We modify a copy of the real _NamespaceInfo so that we are
            # exception-safe: if something goes wrong, we don't end up with a
            # half-changed NamespaceSet.
            ni = self.__lookup_uri(ns_uri)
            new_ni = copy.deepcopy(ni)

            # Reconcile prefixes
            if prefix:
                self.__check_prefix_conflict(ni, prefix)
                new_ni.prefixes.add(prefix)

            self.__merge_schema_locations(new_ni, schema_location)

            # At this point, we have a legit new_ni object.  Now we update
            # the set, ensuring our invariants.  This should replace
            # all instances of the old ni in this set.
            for p in new_ni.prefixes:
                self.__prefix_map[p] = new_ni
            self.__ns_uri_map[new_ni.uri] = new_ni

        else:
            # A brand new namespace.  The incoming prefix should not exist at
            # all in the prefix map.
            if prefix:
                self.__check_prefix_conflict(ns_uri, prefix)

            ni = _NamespaceInfo(ns_uri, prefix, schema_location)
            self.__add_namespaceinfo(ni)

    def remove_namespace(self, ns_uri):
        """Removes the indicated namespace from this set."""
        if not self.contains_namespace(ns_uri):
            return

        ni = self.__ns_uri_map.pop(ns_uri)
        for prefix in ni.prefixes:
            del self.__prefix_map[prefix]

    def add_prefix(self, ns_uri, prefix, set_as_preferred=False):
        """Adds prefix for the given namespace URI.  The namespace must already
        exist in this set.  If set_as_preferred is True, also set this
        namespace as the preferred one.

        ``prefix`` must be non-None; a default preference can't be set this way.
        See :meth:`set_preferred_prefix_for_namespace` for that.

        Args:
            ns_uri (str): The namespace URI to add the prefix to
            prefix (str): The prefix to add (not None)
            set_as_preferred (bool): Whether to set the new prefix as preferred

        Raises:
            NamespaceNotFoundError: If namespace ``ns_uri`` isn't in this set
        """
        assert prefix

        ni = self.__lookup_uri(ns_uri)

        self.__check_prefix_conflict(ni, prefix)
        ni.prefixes.add(prefix)
        self.__prefix_map[prefix] = ni

        if set_as_preferred:
            ni.preferred_prefix = prefix

    def get_prefixes(self, ns_uri):
        """Gets (a copy of) the prefix set for the given namespace."""
        ni = self.__lookup_uri(ns_uri)
        return ni.prefixes.copy()

    def prefix_iter(self, ns_uri):
        """Gets an iterator over the prefixes for the given namespace."""
        ni = self.__lookup_uri(ns_uri)
        return iter(ni.prefixes)

    def remove_prefix(self, prefix):
        """Removes prefix from this set.  This is a no-op if the prefix
        doesn't exist in it.
        """
        if prefix not in self.__prefix_map:
            return

        ni = self.__lookup_prefix(prefix)
        ni.prefixes.discard(prefix)
        del self.__prefix_map[prefix]

        # If we removed the preferred prefix, find a new one.
        if ni.preferred_prefix == prefix:
            ni.preferred_prefix = next(iter(ni.prefixes), None)

    def get_schema_location(self, ns_uri):
        """Gets the schema location URI for the given namespace.

        Args:
            ns_uri (str): The namespace URI whose schema location is needed.

        Returns:
            The schema location, or None if none has been set.

        Raises:
            NamespaceNotFoundError: If the given namespace isn't in this set.
        """
        ni = self.__lookup_uri(ns_uri)
        return ni.schema_location

    def set_schema_location(self, ns_uri, schema_location, replace=False):
        """Sets the schema location of the given namespace.

        If ``replace`` is ``True``, then any existing schema location is
        replaced.  Otherwise, if the schema location is already set to a
        different value, an exception is raised.  If the schema location is set
        to None, it is effectively erased from this set (this is not considered
        "replacement".)

        Args:
            ns_uri (str): The namespace whose schema location is to be set
            schema_location (str): The schema location URI to set, or None
            replace (bool): Whether to replace any existing schema location

        Raises:
            NamespaceNotFoundError: If the given namespace isn't in this set.
            ConflictingSchemaLocationError: If replace is False,
                schema_location is not None, and the namespace already has a
                different schema location in this set.
        """
        ni = self.__lookup_uri(ns_uri)

        if ni.schema_location == schema_location:
            return
        elif replace or ni.schema_location is None:
            ni.schema_location = schema_location
        elif schema_location is None:
            ni.schema_location = None  # Not considered "replacement".
        else:
            raise ConflictingSchemaLocationError(ns_uri, ni.schema_location, schema_location)

    def get_xmlns_string(self, ns_uris=None, sort=False,
                         preferred_prefixes_only=True, delim="\n"):
        """Generates XML namespace declarations for namespaces in this
        set.  It must be suitable for use in an actual XML document,
        so an exception is raised if this can't be done, e.g. if it would
        have more than one default namespace declaration.

        If ``preferred_prefixes_only`` is ``True`` and a namespace's prefix
        preference is to be a default namespace, a default declaration will
        be used if possible.  If that's not possible, a prefix will be
        chosen (is this a good idea?).  If a default declaration can't be used
        and no other prefixes are defined, an exception is raised.

        Args:
            ns_uris (iterable): If non-None, it should be an iterable over
                namespace URIs.  Only the given namespaces will occur in the
                returned string.  If None, all namespace are included.
            sort (bool): If True, the string is constructed from URIs in sorted
                order.
            preferred_prefixes_only (bool): Whether to include only the
                preferred prefix or all of them, for each namespace.
            delim (str): The delimiter to use between namespace declarations.
                Should be some kind of whitespace.

        Returns:
            str: A string in the following format:
                ``xmlns:foo="bar"<delim>xmlns:foo2="bar2"<delim>...``

        Raises:
            NamespaceNotFoundError: If ``ns_uris`` is given and contains any
                URIs not in this set.
            TooManyDefaultNamespacesError: If too many namespaces didn't have
                a prefix.  The algorithm is very simple for deciding whose
                default preference is honored: the first default preference
                encountered gets to be default.  Any subsequent namespaces
                without any prefixes will cause this error.
        """
        if ns_uris is None:
            ns_uris = self.namespace_uris

        if sort:
            ns_uris = sorted(ns_uris)

        have_default  = False  # Flag for default xmlns entry.
        xmlns_entries = []     # Stores all the xmlns:prefix=uri entries.

        for ns_uri in ns_uris:
            ni = self.__lookup_uri(ns_uri)

            if preferred_prefixes_only:
                if ni.preferred_prefix is not None:
                    xmlns = 'xmlns:{0.preferred_prefix}="{0.uri}"'.format(ni)
                    xmlns_entries.append(xmlns)
            else:
                xmlns = 'xmlns:{0}="{1.uri}"'
                xmlns_entries.extend(xmlns.format(prefix, ni) for prefix in ni.prefixes)

            if ni.preferred_prefix is None:
                if have_default:
                    # Already have a default namespace; try to choose a prefix
                    # for this one from the set of registered prefixes.
                    if len(ni.prefixes) == 0:
                        raise TooManyDefaultNamespacesError(ni.uri)
                    elif preferred_prefixes_only:
                        prefix = next(iter(ni.prefixes))
                        xmlns  = 'xmlns:{0}="{1.uri}"'.format(prefix, ni)
                        xmlns_entries.append(xmlns)

                    # else, we already declared some prefixes for this
                    # namespace, so don't worry about our inability to use this
                    # as a default namespace.
                else:
                    xmlns = 'xmlns="{0.uri}"'.format(ni)
                    xmlns_entries.append(xmlns)
                    have_default = True

        xmlns_str = delim.join(xmlns_entries) + delim
        return xmlns_str

    def get_schemaloc_string(self, ns_uris=None, sort=False, delim="\n"):
        """Constructs and returns a schemalocation attribute.  If no
        namespaces in this set have any schema locations defined, returns
        an empty string.

        Args:
            ns_uris (iterable): The namespaces to include in the constructed
                attribute value.  If None, all are included.
            sort (bool): Whether the sort the namespace URIs.
            delim (str): The delimiter to use between namespace/schemaloc
                *pairs*.

        Returns:
            str: A schemalocation attribute in the format:
                ``xsi:schemaLocation="nsuri schemaloc<delim>nsuri2 schemaloc2<delim>..."``

        """
        if not ns_uris:
            ns_uris = six.iterkeys(self.__ns_uri_map)

        if sort:
            ns_uris = sorted(ns_uris)

        schemalocs = []

        for ns_uri in ns_uris:
            ni = self.__lookup_uri(ns_uri)

            if ni.schema_location:
                schemalocs.append("{0.uri} {0.schema_location}".format(ni))

        if not schemalocs:
            return ""

        return 'xsi:schemaLocation="{0}"'.format(delim.join(schemalocs))

    def get_uri_prefix_map(self):
        """Constructs and returns a map from namespace URI to prefix,
        representing all namespaces in this set.  The prefix chosen for each
        namespace is its preferred prefix if it's not None.  If the preferred
        prefix is None, one is chosen from the set of registered
        prefixes.  In the latter situation, if no prefixes are registered,
        an exception is raised.
        """
        mapping = {}
        
        for ni in six.itervalues(self.__ns_uri_map):
            if ni.preferred_prefix:
                mapping[ni.uri] = ni.preferred_prefix
            elif len(ni.prefixes) > 0:
                mapping[ni.uri] = next(iter(ni.prefixes))
            else:
                # The reason I don't let any namespace map to None here is that
                # I don't think generateDS supports it.  It requires prefixes
                # for all namespaces.
                raise NoPrefixesError(ni.uri)

        return mapping

    def get_prefix_uri_map(self):
        """Constructs and returns a map from to prefix to namespace URI,
        representing all namespaces in this set.  The prefix chosen for each
        namespace is its preferred prefix if it's not None.  If the preferred
        prefix is None, one is chosen from the set of registered
        prefixes.  In the latter situation, if no prefixes are registered,
        an exception is raised.
        """
        mapping = {}

        for ni in six.itervalues(self.__ns_uri_map):
            if ni.preferred_prefix:
                mapping[ni.preferred_prefix] = ni.uri
            elif len(ni.prefixes) > 0:
                prefix = next(iter(ni.prefixes))
                mapping[prefix] = ni.uri
            else:
                raise NoPrefixesError(ni.uri)

        return mapping

    def get_uri_schemaloc_map(self):
        """Constructs and returns a map from namespace URI to schema location
        URI.  Namespaces without schema locations are excluded."""
        mapping = {}

        for ni in six.itervalues(self.__ns_uri_map):
            if ni.schema_location:
                mapping[ni.uri] = ni.schema_location

        return mapping

    @property
    def namespace_uris(self):
        """A generator over the namespace URIs stored in this set."""
        for uri in six.iterkeys(self.__ns_uri_map):
            yield uri

    def subset(self, ns_uris):
        """Return a subset of this NamespaceSet containing only data for the
        given namespaces.

        Args:
            ns_uris (iterable): An iterable of namespace URIs which select the
                namespaces for the subset.

        Returns:
            The subset

        Raises:
            NamespaceNotFoundError: If any namespace URIs in `ns_uris` don't
                match any namespaces in this set.
        """
        sub_ns = NamespaceSet()

        for ns_uri in ns_uris:
            ni = self.__lookup_uri(ns_uri)
            new_ni = copy.deepcopy(ni)

            # We should be able to reach into details of our own
            # implementation on another obj, right??  This makes the subset
            # operation faster.  We can set up the innards directly from a
            # cloned _NamespaceInfo.
            sub_ns._NamespaceSet__add_namespaceinfo(new_ni)

        return sub_ns

    def import_from(self, other_ns, replace=False):
        """Imports namespaces into this set, from other_ns.

        Args:
            other_ns (NamespaceSet): The set to import from
            replace (bool): If a namespace exists in both sets, do we replace
                our data with other_ns's data?  We could get fancy and define
                some merge strategies, but for now, this is very simple.  It's
                either do nothing, or wholesale replacement.  There is no
                merging.

        Raises:
            DuplicatePrefixError: If the other NamespaceSet is mapping any
                prefixes incompatibly with respect to this set.
        """
        for other_ns_uri in other_ns.namespace_uris:
            ni = self.__ns_uri_map.get(other_ns_uri)

            if ni is None:
                other_ni = other_ns._NamespaceSet__ns_uri_map[other_ns_uri]

                # Gotta make sure that the other set isn't mapping its prefixes
                # incompatibly with respect to this set.
                for other_prefix in other_ni.prefixes:
                    self.__check_prefix_conflict(other_ns_uri, other_prefix)

                cloned_ni = copy.deepcopy(other_ni)
                self.__add_namespaceinfo(cloned_ni)
            elif replace:
                other_ni = other_ns._NamespaceSet__ns_uri_map[other_ns_uri]
                for other_prefix in other_ni.prefixes:
                    self.__check_prefix_conflict(ni, other_prefix)

                cloned_ni = copy.deepcopy(other_ni)
                self.remove_namespace(other_ns_uri)
                self.__add_namespaceinfo(cloned_ni)
            else:
                continue  # TODO (bworrell): Should this log/warn/do anything else?

    def assert_valid(self):
        """For debugging; does some sanity checks on this set.  Raises
        InvalidNamespaceSetError if this namespace set is invalid.  Otherwise,
        raises/returns nothing."""

        for ns_uri, ni in six.iteritems(self.__ns_uri_map):
            if not ni.uri:
                raise InvalidNamespaceSetError(
                    "URI not set in _NamespaceInfo (id={0}):\n{1}".format(
                        id(ni), ni
                    )
                )

            if ns_uri != ni.uri:
                raise InvalidNamespaceSetError(
                    "URI mismatch in dict ({0}) and _NamespaceInfo ({1})".format(
                        ns_uri, ni.uri
                    )
                )

            if (ni.preferred_prefix is not None and
               ni.preferred_prefix not in ni.prefixes):
                raise InvalidNamespaceSetError(
                    "Namespace {0.uri}: preferred prefix " \
                    '"{0.preferred_prefix}" not in prefixes ' \
                    "{0.prefixes}".format(ni)
                )

            for prefix in ni.prefixes:
                if not prefix:
                    raise InvalidNamespaceSetError(
                        "Namespace {0.uri}: empty value in prefix " \
                        "set: {0.prefixes}".format(ni)
                    )
                other_ni = self.__prefix_map.get(prefix)
                if other_ni is None:
                    raise InvalidNamespaceSetError(
                        'Namespace {0.uri}: prefix "{1}" not in ' \
                        'prefix map'.format(ni, prefix)
                    )
                if other_ni is not ni:
                    raise InvalidNamespaceSetError(
                        'Namespace {0.uri}: prefix "{1}" maps to ' \
                        'wrong _NamespaceInfo (id={2}, uri={3.uri})'.format(
                            ni, prefix, id(other_ni), other_ni
                        )
                    )

        if None in self.__prefix_map:
            # None can be a preferred prefix, but should not be in the
            # prefix map.
            raise InvalidNamespaceSetError("None is in prefix map!")

    def is_valid(self):
        """Does some sanity checks on this set.  Returns True if the set is
        valid, False otherwise."""
        try:
            self.assert_valid()
        except InvalidNamespaceSetError:
            return False

        return True

    def __len__(self):
        """Return the number of namespaces in this set."""
        return len(self.__ns_uri_map)

    def __eq__(self, other):
        """Test two namespaces for equality.  For them to be equal, they must
        contain the same namespaces, and for each namespace, their prefixes,
        schema location, and preferred prefix must also be the same."""

        if type(other) != type(self):
            return False
        elif len(other) != len(self):
            return False
        else:
            return self.__ns_uri_map == other._NamespaceSet__ns_uri_map

    def __ne__(self, other):
        """Python2 apparently needs this; python3 has a suitable default
        which delegates to __eq__.
        """
        return not self.__eq__(other)

    def __str__(self):
        """for debugging"""
        return "\n\n".join(str(v) for v in six.itervalues(self.__ns_uri_map))


__ALL_NAMESPACES = NamespaceSet()


def register_namespace(namespace):
    """Register a new Namespace with the global NamespaceSet."""

    __ALL_NAMESPACES.add_namespace(namespace)


def lookup_name(ns_uri):
    """Return the preferred prefix for the input namespace uri."""
    return __ALL_NAMESPACES.preferred_prefix_for_namespace(ns_uri)


def lookup_prefix(prefix):
    """Return the namespace uri that is mapped to the input prefix."""
    return __ALL_NAMESPACES.namespace_for_prefix(prefix)


def make_namespace_subset_from_uris(ns_uris):
    """Creates a subset of the global NamespaceSet containing info only for
    the given namespaces."""
    return __ALL_NAMESPACES.subset(ns_uris)


def get_full_ns_map():
    """Return a name: prefix mapping for all registered Namespaces."""
    return __ALL_NAMESPACES.get_uri_prefix_map()


def get_full_prefix_map():
    """Return a prefix: name mapping for all registered Namespaces."""
    return __ALL_NAMESPACES.get_prefix_uri_map()


def get_full_schemaloc_map():
    """Return a name: schemalocation mapping for all registered Namespaces."""
    return __ALL_NAMESPACES.get_uri_schemaloc_map()


def get_xmlns_string(ns_uris=None, sort=False):
    """Build a string with 'xmlns' definitions for every namespace in ns_uris.
    If ns_uris is None, all namespaces are included.

    Args:
        ns_uris (iterable): the namespace URIs, or None
    """
    return __ALL_NAMESPACES.get_xmlns_string(ns_uris, sort)


def get_schemaloc_string(ns_uris=None, sort=False):
    """Build a "schemaLocation" string for every namespace in ns_uris.
    If ns_uris is None, all namespaces are included.

    Args:
        ns_uris (iterable): the namespace URIs, or None
    """
    return __ALL_NAMESPACES.get_schemaloc_string(ns_uris, sort)


NS_XLINK = Namespace('http://www.w3.org/1999/xlink', 'xlink', '')
NS_XML_DSIG = Namespace('http://www.w3.org/2000/09/xmldsig#', 'ds', '')
NS_XML_SCHEMA = Namespace('http://www.w3.org/2001/XMLSchema', 'xs', '')
NS_XML_SCHEMA_INSTANCE = Namespace('http://www.w3.org/2001/XMLSchema-instance', 'xsi', '')

XML_NAMESPACES = NamespaceSet()

# Magic to automatically register all Namespaces defined in this module.
for k, v in dict(globals()).items():
    if k.startswith('NS_'):
        register_namespace(v)
        XML_NAMESPACES.add_namespace(v)
